import pygame

from gummworld2 import State, Vec2d, data
from gummworld2.geometry import point_on_circumference


def slot_to_screen(slot):
    screen_width,screen_height = State.screen.rect.size
    y = screen_height - 32
    x = slot * 32
    while x+32 > screen_width:
        y -= 32
        slot -= screen_width / 32
        x = slot * 32
    return x,y


WORDS = {
    1000 : '1K',
    10000 : '10K',
    100000 : '100K',
    1000000 : '1M',
    10000000 : '10M',
    100000000 : '100M',
    1000000000 : '1B',
}


class Sprite(object):
    
    def __init__(self, image, slot):
        self.image = image
        self.rect = image.get_rect(center=State.camera.screen_center)
        self.slot = slot
        self.goto = Vec2d(slot_to_screen(slot))
        self.pos = Vec2d(self.rect.topleft)
        self.home = False
    
    def update(self, dt):
        if self.home:
            return
        pos = self.pos
        goto = self.goto
        if pos != goto:
            dx = goto.x - pos.x
            dy = goto.y - pos.y
            if abs(dx) <= 1:
                x = goto.x
            else:
                dx *= 0.1
                x = pos.x + dx
            if abs(dy) <= 1:
                y = goto.y
            else:
                dy *= 0.1
                y = pos.y + dy
            self.pos[:] = x,y
            self.rect.topleft = round(x),round(y)
        else:
            self.home = True


class Score(object):
    
    def __init__(self):
        self.tallies = dict(
            cheeseballs_fired=0,
            cheeseballs_missed=0,
            cheeseballs_hit=0,
            moons_harvested=0,
            moons_lost=0,
            moons_smashed=0,
            tethers_fired=0,
            tethers_hit=0,
            tethers_missed=0)
        
        self.achievements = []
        self.sprites = []
#        self.cleanup_slots = []
        
        # Achievement brokers
        CheeseballsFired(self)
        MoonsHarvested(self)
        #
        Sharpshooter(self)
        Expert(self)
        Cheeseeye(self)
        #
        Cowpoke(self)
        Rustler(self)
        Wrangler(self)
        #
        Picker(self)
        Gatherer(self)
        Harvester(self)
        #
        PunkinSmasher(self)
    
    def add(self, achievement):
        self.achievements.append(achievement)
    
    def remove(self, achievement):
        if achievement in self.achievements:
            self.achievements.remove(achievement)
#            self.cleanup_slots.append(achievement.slot)
    
    def add_sprite(self, sprite):
        self.sprites.append(sprite)
    
    def achievements_in_slot(self, slot):
        return [a for a in self.achievements if a.slot == slot]
    
    def sprites_in_slot(self, slot):
        return [s for s in self.sprites if s.slot == slot]
    
    def cleanup_slot(self, slot):
        sprites = self.sprites_in_slot(slot)
        home = [s.home for s in sprites]
        if len(sprites) > 1:
            if all(home):
                self.sprites.remove(sprites[0])
#                self.cleanup_slots.remove(slot)
#        else:
#            self.cleanup_slots.remove(slot)
    
    def bump(self, attr):
        self.tallies[attr] += 1
    
    def update(self, dt):
        for achievement in self.achievements:
            achievement.check()
        for sprite in self.sprites:
            sprite.update(dt)
#        for slot in self.cleanup_slots[:]:
        slots = set([s.slot for s in self.sprites])
        for slot in slots:
            self.cleanup_slot(slot)
    
    def award(self, slot):
        # For testing awards only.
        for a in self.achievements:
            if a.slot == slot:
                a.award()
                break


class Achievement(object):
    
    slot = None
    image_size = 32,32
    font = None
    font_color = None
    next = None
    
    def __init__(self, score):
        self.score = score
        score.add(self)
        if self.font is None:
            font_file = data.filepath('font', 'Vera.ttf')
            self.font = pygame.font.Font(font_file, 8)
            self.font_color = pygame.Color('yellow')
    
    def sync(self):
        pass
    
    def goal_reached(self):
        return False
    
    def check(self):
        if self.goal_reached():
            self.award()
        self.sync()
    
    def award(self):
        image = self.make_image()
        if image:
            self.score.add_sprite(Sprite(image, self.slot))
        self.complete()
        if self.next:
            self.next(self.score)
    
    def complete(self):
        self.score.remove(self)


class CheeseballsFired(Achievement):
    
    slot = 0
    cheeseballs_needed = 100
    label = '100'
    name = 'CheeseballsFired'
    
    def __init__(self, score):
        super(CheeseballsFired, self).__init__(score)
    
    def goal_reached(self):
        return self.score.tallies['cheeseballs_fired'] >= self.cheeseballs_needed
    
    def complete(self):
        if self.cheeseballs_needed >= 1000:
            self.cheeseballs_needed *= 10
        else:
            self.cheeseballs_needed += 100
        self.label = str(self.cheeseballs_needed)
    
    def make_image(self):
        if __debug__: print self.name, 'earned', self.cheeseballs_needed
        image = pygame.Surface(self.image_size)
        rect = image.get_rect()
        pygame.draw.rect(image, self.font_color, rect, 1)
        # label
        label = WORDS.get(self.cheeseballs_needed, self.label)
        text = self.font.render(label, 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        # cheeseball
        text_rect.bottom = rect.bottom - 3
        pygame.draw.circle(image, pygame.Color('yellow2'), text_rect.center, 4)
        
        return image

class MoonsHarvested(Achievement):
    
    slot = 1
    moons_needed = 10
    label = '10'
    name = 'MoonsHarvested'
    
    def __init__(self, score):
        super(MoonsHarvested, self).__init__(score)
    
    def goal_reached(self):
        return self.score.tallies['moons_harvested'] >= self.moons_needed
    
    def complete(self):
        if self.moons_needed >= 1000:
            self.moons_needed *= 10
        else:
            self.moons_needed += 10
        self.label = str(self.moons_needed)
    
    def make_image(self):
        if __debug__: print self.name, 'earned', self.label
        image = pygame.Surface(self.image_size)
        rect = image.get_rect()
        pygame.draw.rect(image, self.font_color, rect, 1)
        # label
        label = WORDS.get(self.moons_needed, self.label)
        text = self.font.render(label, 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        # moon
        text_rect.bottom = rect.bottom - 3 - 4
        pygame.draw.circle(image, pygame.Color('lightyellow'), text_rect.center, 7)
        # crescent
        radius = 7.0 / 2.0
        shift = Vec2d(text_rect.center)
        fact = 3.0/4.0
        shift.x += radius * fact
        shift.y -= radius * fact
        o = int(round(shift.x)),int(round(shift.y))
        r = int(round(radius))
        pygame.draw.circle(image, (0,0,0), o, r)
        
        return image


class Sharpshooter(Achievement):
    
    slot = 2
    accuracy_lo = 60.0
    accuracy_hi = 80.0
    num_shots = 100
    label = 'X'
    
    def __init__(self, score):
        super(Sharpshooter, self).__init__(score)
        self.name = 'Sharpshooter'
        self.missed = 0
        self.hit = 0
        self.accuracy = 0
    
    def goal_reached(self):
        t = self.score.tallies
        hit = t['cheeseballs_hit'] - self.hit
        missed = t['cheeseballs_missed'] - self.missed
        if hit != 0:
            n = hit - missed
            if n < 0:
                n = 0
            self.accuracy = round(100.0*n/hit, 2)
        if t['cheeseballs_hit'] < self.num_shots:
            return False
        if self.accuracy >= self.accuracy_lo and \
                self.accuracy < self.accuracy_hi:
            return True
        else:
            self.complete()
            return False
    
    def complete(self):
        self.num_shots += 100
        t = self.score.tallies
        self.missed = t['cheeseballs_missed']
        self.hit = t['cheeseballs_hit']
    
    def make_image(self):
        if __debug__: print self.name, 'earned', self.label
        image = pygame.Surface(self.image_size)
        rect = image.get_rect()
        pygame.draw.rect(image, self.font_color, rect, 1)
        
        text = self.font.render(str(self.label), 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        
        text_rect.bottom = rect.bottom - 3
        pygame.draw.circle(image, self.font_color, text_rect.center, 7)
        
        p1 = Vec2d(text_rect.center)
        p2 = Vec2d(text_rect.center)
        p1.y -= 4
        p2.y += 4
        pygame.draw.line(image, (255,0,0), p1, p2)
        p1 = Vec2d(text_rect.center)
        p2 = Vec2d(text_rect.center)
        p1.x -= 4
        p2.x += 4
        pygame.draw.line(image, (255,0,0), p1, p2)
        
        return image


class Expert(Sharpshooter):
    
    slot = 2
    accuracy_lo = 80.0
    accuracy_hi = 90.0
    label = 'XX'
    
    def __init__(self, score):
        Sharpshooter.__init__(self, score)
        self.name = 'Expert'


class Cheeseeye(Sharpshooter):
    
    slot = 2
    accuracy_lo = 90.0
    accuracy_hi = 101.0
    label = 'XXX'
    
    def __init__(self, score):
        Sharpshooter.__init__(self, score)
        self.name = 'Cheeseeye'


class Cowpoke(Achievement):
    
    slot = 3
    accuracy_lo = 60.0
    accuracy_hi = 80.0
    num_shots = 10
    label = 'X'
    
    def __init__(self, score):
        super(Cowpoke, self).__init__(score)
        self.name = 'Cowpoke'
        self.missed = 0
        self.hit = 0
        self.accuracy = 0
    
    def goal_reached(self):
        t = self.score.tallies
        hit = t['tethers_hit'] - self.hit
        missed = t['tethers_missed'] - self.missed
        if hit != 0:
            n = hit - missed
            if n < 0:
                n = 0
            self.accuracy = round(100.0*n/hit, 2)
        if t['tethers_hit'] < self.num_shots:
            return False
        if self.accuracy >= self.accuracy_lo and \
                self.accuracy < self.accuracy_hi:
            return True
        else:
            self.complete()
            return False
    
    def complete(self):
        self.num_shots += 10
        t = self.score.tallies
        self.missed = t['tethers_missed']
        self.hit = t['tethers_hit']
    
    def make_image(self):
        if __debug__: print self.name,'earned', self.label
        name = data.filepath('image', 'cowpoke.png')
        image = pygame.image.load(name)
        rect = image.get_rect()
        # label
        text = self.font.render(str(self.label), 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        
        return image


class Rustler(Cowpoke):
    
    slot = 3
    accuracy_lo = 90.0
    accuracy_hi = 101.0
    label = 'XX'
    
    def __init__(self, score):
        Cowpoke.__init__(self, score)
        self.name = 'Rustler'


class Wrangler(Cowpoke):
    
    slot = 3
    accuracy_lo = 90.0
    accuracy_hi = 101.0
    label = 'XXX'
    
    def __init__(self, score):
        Cowpoke.__init__(self, score)
        self.name = 'Wrangler'




class Picker(Achievement):
    
    slot = 4
    accuracy_lo = 60.0
    accuracy_hi = 80.0
    num_shots = 10
    label = 'O'
    
    def __init__(self, score):
        super(Picker, self).__init__(score)
        self.name = 'Picker'
        self.harvested = 0
        self.lost = 0
        self.accuracy = 0
    
    def goal_reached(self):
        t = self.score.tallies
        harvested = t['moons_harvested'] - self.harvested
        lost = t['moons_lost'] - self.lost
        total = harvested + lost
        if total != 0:
            self.accuracy = round(100.0*harvested/total, 2)
        if total < self.num_shots:
            return False
        if self.accuracy >= self.accuracy_lo and \
                self.accuracy < self.accuracy_hi:
            return True
        else:
            self.complete()
            return False
    
    def complete(self):
        self.num_shots += 10
        t = self.score.tallies
        self.lost = t['moons_lost']
        self.harvested = t['moons_harvested']
    
    def make_image(self):
        if __debug__: print self.name,'earned', self.label
        image = pygame.Surface(self.image_size)
        rect = image.get_rect()
        pygame.draw.rect(image, self.font_color, rect, 1)
        
        text = self.font.render(str(self.label), 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        
        text_rect.bottom = rect.bottom - 3
        pygame.draw.circle(image, self.font_color, text_rect.center, 7)
        
        p1 = Vec2d(text_rect.center)
        p2 = Vec2d(text_rect.center)
        p1.y -= 4
        p2.y += 4
        pygame.draw.line(image, (255,0,0), p1, p2)
        p1 = Vec2d(text_rect.center)
        p2 = Vec2d(text_rect.center)
        p1.x -= 4
        p2.x += 4
        pygame.draw.line(image, (255,0,0), p1, p2)
        
        return image


class Gatherer(Picker):
    
    slot = 4
    accuracy_lo = 90.0
    accuracy_hi = 101.0
    label = 'OO'
    
    def __init__(self, score):
        Picker.__init__(self, score)
        self.name = 'Rustler'


class Harvester(Picker):
    
    slot = 4
    accuracy_lo = 90.0
    accuracy_hi = 101.0
    label = 'OOO'
    
    def __init__(self, score):
        Picker.__init__(self, score)
        self.name = 'Wrangler'


class PunkinSmasher(Achievement):
    
    slot = 5
    moons_needed = 50
    label = '50'
    name = 'PunkinSmasher'
    
    def __init__(self, score):
        super(PunkinSmasher, self).__init__(score)
    
    def goal_reached(self):
        return self.score.tallies['moons_smashed'] >= self.moons_needed
    
    def completed(self):
        if self.moons_needed >= 1000:
            self.moons_needed *= 10
        else:
            self.moons_needed += 50
        self.label = str(self.moons_needed)
    
    def make_image(self):
        if __debug__: print self.name, 'earned', self.label
        image = pygame.Surface(self.image_size)
        rect = image.get_rect()
        pygame.draw.rect(image, self.font_color, rect, 1)
        # label
        label = WORDS.get(self.moons_needed, self.label)
        text = self.font.render(self.label, 1, self.font_color)
        text_rect = text.get_rect(centerx=rect.centerx, y=3)
        image.blit(text, text_rect)
        # moon
        text_rect.bottom = rect.bottom - 3 - 4
        pygame.draw.circle(image, pygame.Color('lightyellow'), text_rect.center, 7)
        # crescent
        radius = 7.0 / 2.0
        shift = Vec2d(text_rect.center)
        fact = 3.0/4.0
        shift.x += radius * fact
        shift.y -= radius * fact
        o = int(round(shift.x)),int(round(shift.y))
        r = int(round(radius))
        pygame.draw.circle(image, (0,0,0), o, r)
        # bar sinister
        o = text_rect.center
        pygame.draw.circle(image, pygame.Color('red'), text_rect.center, 7, 1)
        p1 = point_on_circumference(o, radius, 45)
        p2 = point_on_circumference(o, radius, 225)
        pygame.draw.line(image, (255,0,0), p1, p2)
        
        return image
