"""game.py - a context for running a game session or level

This file is part of Gummworld2.

Gummworld2 is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Gummworld2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.
"""

import random

import pygame
from pygame.locals import (
    Color, Rect,
    K_ESCAPE, K_SPACE, K_RETURN, K_g, K_m,
    ACTIVEEVENT,
    BLEND_RGBA_SUB, RLEACCEL,
)

import gummworld2
from gummworld2 import State, Vec2d, SpatialHash, HUD, Statf
from gummworld2 import model
from gummworld2.geometry import (
    LineGeometry, CircleGeometry, RectGeometry,
    line_intersects_rect, circle_intersects_line, line_intersects_line,
    point_on_circumference, distance, interpolant_of_line,
)

import settings, pixelperfect, sound, achievements


def ir(f):
    return int(round(f))

def ir2(ff):
    return int(round(ff[0])),int(round(ff[1]))


class Home(RectGeometry):
    
    def __init__(self):
        x = settings.resolution[0] * 1/2
        y = settings.ground_level - 20
        w = h = 20
        super(Home, self).__init__(x,y, w,h)
        self.color = Color('orange')


class Ground(LineGeometry):
    
    def __init__(self):
        x = 0
        y = settings.ground_level
        w = settings.resolution[0]
        h = y
        super(Ground, self).__init__(x,y, w,h)
        self.color = Color('green')


class Chute(object):
    
    def __init__(self, origin, radius, angle):
        self.origin = origin
        self.radius = radius
        self.angle = angle
        self.color = Color('white')
    
    def _get_muzzle_pos(self):
        return point_on_circumference(self.origin, self.radius, self.angle)
    muzzle_pos = property(_get_muzzle_pos)
    
    def _get_points(self):
        return self.origin, point_on_circumference(
            self.origin, self.radius, self.angle)
    points = property(_get_points)


class CheeseBall(CircleGeometry):
    
    hitmask = None
    radius = 4.0
    color = None
    
    def __init__(self, position, angle):
        super(CheeseBall, self).__init__(position, self.radius)
        if self.color is None:
            self.color = Color('yellow2')
        w = h = self.radius * 2
        self.image = pygame.Surface((w,h))
        center = (self.radius, self.radius)
        pygame.draw.circle(self.image, self.color, ir2(center), ir(self.radius))
        self.image.set_colorkey((0,0,0))
        speed = settings.cheese_speed * settings.dt
        x1,y1 = position
        x2,y2 = point_on_circumference(position, speed, angle)
        self.speed = Vec2d(x2-x1,y2-y1)
        if self.hitmask is None:
            self.hitmask = pixelperfect.get_colorkey_hitmask(self.image, self.rect)
    
    def update(self, dt):
        self.position += self.speed
        self.speed.y += settings.G * dt


class Moon(CircleGeometry):
    
    num_phases = 4
    color = None
    
    def __init__(self, origin, radius, phase):
        if self.color is None:
            self.color = Color('lightyellow')
        im = pygame.Surface((radius*3,radius*3), RLEACCEL)
        r = im.get_rect()
        pygame.draw.circle(im, self.color, r.center, ir(radius))
        shift = Vec2d(r.center)
        fact = float(phase+1)/self.num_phases
        shift.x += r.w/2 * fact
        shift.y -= r.w/2 * fact
        pygame.draw.circle(im, (0,0,0), ir2(shift), ir(radius))
        im.set_colorkey((0,0,0))
        
        self.image = im
        self.image.set_colorkey((0,0,0))
        super(Moon, self).__init__(origin, radius)
        self.rect = self.image.get_rect(center=self.position)
        self.speed = settings.moon_speed * settings.dt
        self.hitmask = pixelperfect.get_colorkey_hitmask(self.image, self.rect)
        self.hits = 0
        self.attached = None
        p = self.position
#        self.position = 200,p[1]
    
    def update(self, dt):
        if not self.attached:
            self.position += (self.speed, 2*self.hits*dt)
    
    def hit(self, other):
        try:
            contact = pixelperfect.check_collision(self, other)
        except AttributeError:
            return False
        if contact:
            p = Vec2d(other.rect.topleft) - self.rect.topleft
            self.image.blit(other.image, p)
            self.hits += 1
            return True
        else:
            return False


class Tether(LineGeometry):
    
    color = None
    max_distance = 150.0
    
    def __init__(self, origin, angle):
        x1,y1 = origin
        x2,y2 = point_on_circumference(origin, 40.0, angle)
        super(Tether, self).__init__(x1,y1, x2,y2)
        if self.color is None:
            self.color = Color('violet')
        speed = settings.tether_speed * settings.dt
        x2,y2 = point_on_circumference((x1,y1), speed, angle)
        self.speed = Vec2d(x2-x1,y2-y1)
        self.attached = None
        self.anim = 0
    
    def update(self, dt):
        if self.attached:
            p1 = Vec2d(self.p1)
            p2 = Vec2d(self.p2)
            dx,dy = p2 - p1
            mag = 0.01
            dx *= mag
            dy *= mag
            self.p2 -= Vec2d(dx,dy)
            self.attached.position -= (dx,dy)
        else:
            self.p2 += self.speed
            self.speed.x *= 0.98
            self.speed.y += settings.G * dt
        if self.anim:
            self.update_anim(dt)
        
    def update_anim(self, dt):
        pass
    
    def attach(self, moon):
        moon.attached = self
        self.attached = moon
        p = circle_intersects_line(moon.origin, moon.radius, self.points)
        if p:
            self.p2 = p
        self.anim = 1


class Game(gummworld2.Engine):
    
    def __init__(self):
        res = Vec2d(settings.resolution)
        
        gummworld2.Engine.__init__(self,
            caption=settings.caption,
            resolution=res,
            update_speed=settings.update_speed,
            frame_speed=settings.update_speed,  # sic
            camera_target=model.Object(res/2),
        )
        State.clock.paused_callback = self.paused
        self.do_focus = False
        self.space = SpatialHash(State.screen.rect, 64)
        
        sound.song_volume(settings.music_volume)
        sound.load_songs()
        
        # Ground
        self.ground = Ground()
        
        # Home
        self.home = Home()
        
        # Chute
        r = self.home.rect
        self.chute = Chute((r.centerx, r.bottom-1), 40.0, 0.0)
        
        self.mouse_buttons = [None,False,False,False]
        self.cheese_cooldown = 0.0
        self.tether_cooldown = 0.0
        self.song_cooldown = 0.0
        self.which_moon = 0
        
        self.harvested = 0
        self.score = achievements.Score()
        
        State.hud = HUD()
        next_pos = State.hud.next_pos
        State.hud.add('FPS', Statf(next_pos(),
            'FPS %d', callback=lambda:State.clock.fps))
#        State.hud.add('angle', Statf(next_pos(),
#            'Angle %0.2f', callback=lambda:self.chute.angle, interval=0.1))
        def cheese():
            t = self.score.tallies
            f = t['cheeseballs_fired']
            m = t['cheeseballs_missed']
            h = t['cheeseballs_hit']
            aa = self.score.achievements_in_slot(2)
            return '{0}/{1}/{2}/{3:0.1f}'.format(f,h,m,aa[0].accuracy)
        State.hud.add('cheese', Statf(next_pos(),
            'Cheeseballs F/H/M/A %s', callback=cheese, interval=0.1))
        def moon():
            t = self.score.tallies
            h = t['moons_harvested']
            l = t['moons_lost']
            s = t['moons_smashed']
            aa = self.score.achievements_in_slot(3)
            return '{0}/{1}/{2}/{3:0.1f}'.format(h,l,s,aa[0].accuracy)
        State.hud.add('moons', Statf(next_pos(),
            'Moons H/L/S/A %s', callback=moon, interval=0.1))
        def tether():
            t = self.score.tallies
            f = t['tethers_fired']
            h = t['tethers_hit']
            m = t['tethers_missed']
            aa = self.score.achievements_in_slot(3)
            return '{0}/{1}/{2}/{3:0.1f}'.format(f,h,m,aa[0].accuracy)
        State.hud.add('tethers', Statf(next_pos(),
            'Lassoos F/H/M/A %s', callback=tether, interval=0.1))
        
        State.clock.schedule_interval(self.make_moon, interval=4.0)
        
        pygame.event.set_grab(1)
        pygame.mouse.set_visible(0)
    
    def fire_cheese(self, dt):
        self.cheese_cooldown -= dt
        if self.cheese_cooldown > 0.0:
            return
        if self.mouse_buttons[1]:
            cheese_ball = CheeseBall(self.chute.muzzle_pos, self.chute.angle)
            self.cheese_cooldown = settings.cheese_cooldown
            self.space.add(cheese_ball)
            self.score.bump('cheeseballs_fired')
    
    def fire_tether(self, dt):
        self.tether_cooldown -= dt
        if self.tether_cooldown > 0.0:
            return
        if self.mouse_buttons[3]:
            tether = Tether(self.chute.origin, self.chute.angle)
            self.tether_cooldown = settings.tether_cooldown
            self.space.add(tether)
            self.mouse_buttons[3] = False
            self.score.bump('tethers_fired')
   
    def make_moon(self, *args):
        radius = random.randrange(*settings.moon_size)
        x = -radius
        y = random.randrange(*settings.moon_y)
        moon = Moon((x,y), radius, random.choice(settings.moon_freq))
        self.space.add(moon)
    
    def rack_em_up(self, moon):
        self.score.bump('moons_harvested')
    
    def update(self, dt):
        State.camera.update()
        self.fire_cheese(dt)
        self.fire_tether(dt)
        self.update_space(dt)
        self.update_harvest()
        self.score.update(dt)
        self.update_music(dt)
        if settings.show_hud:
            State.hud.update(dt)
    
    def update_space(self, dt):
        screen_rect = State.screen.rect
        colliderect = screen_rect.colliderect
        space = self.space
        ground_level = settings.ground_level
        for o in space.objects:
            o.update(dt)
            try:
                if o in space:
                    space.add(o)
            except IndexError:
                # Unceremoniously dump anything outside the space.
                if isinstance(o, CheeseBall):
                    self.score.bump('cheeseballs_missed')
                elif isinstance(o, Moon):
                    self.score.bump('moons_lost')
                space.remove(o)
                continue
            if isinstance(o, CheeseBall):
                r = o.rect
                if r.right < screen_rect.left or r.top > ground_level:
                    space.remove(o)
                    self.score.bump('cheeseballs_missed')
            elif isinstance(o, Moon):
                r = o.rect
                if r.left > screen_rect.right:
                    space.remove(o)
                    self.score.bump('moons_lost')
                    if o.attached:
                        space.remove(o.attached)
                else:
                    self.update_moon_collisions(o)
            elif isinstance(o, Tether):
                self.update_tether_collisions(o)
    
    def update_moon_collisions(self, moon):
        if moon.attached:
            return
        space = self.space
        collisions = space.collide(moon, CheeseBall)
        for cheeseball in collisions:
            if moon.hit(cheeseball):
                space.remove(cheeseball)
                self.score.bump('cheeseballs_hit')
        if circle_intersects_line(moon.origin, moon.radius, self.ground.points):
            space.remove(moon)
            self.score.bump('moons_smashed')
            self.score.bump('moons_lost')
            if moon.attached:
                space.remove(moon.attached)
    
    def update_tether_collisions(self, tether):
        if tether.attached:
            return
        space = self.space
        if line_intersects_line(tether.points, self.ground.points):
            space.remove(tether)
            self.score.bump('tethers_missed')
            return
        collisions = space.collide(tether, Moon)
        for moon in collisions:
            if moon.attached is None:
                tether.attach(moon)
                self.score.bump('tethers_hit')
                break
    
    def update_harvest(self):
        space = self.space
        collisions = space.collide(self.home, Moon)
        for moon in collisions:
            if moon.attached:
                self.rack_em_up(moon)
                space.remove(moon.attached)
            space.remove(moon)
    
    def update_music(self, dt):
        if sound.song_is_playing():
            if self.song_cooldown < settings.intermission:
                print 'song playing'
            self.song_cooldown = settings.intermission
        elif self.song_cooldown > 0.0:
            if self.song_cooldown == settings.intermission:
                print 'cooling down'
            self.song_cooldown -= dt
        else:
#            print 'next song'
            sound.next_song(dt)
    
    def draw(self, interp):
        self.screen.fill((0,0,0))
        self.draw_fixtures()
        self.draw_objects()
        self.draw_achievements()
        if settings.show_hud:
            State.hud.draw()
        self.screen.flip()
    
    def draw_fixtures(self):
        surf = State.screen.surface
        draw_line = pygame.draw.aaline
        draw_lines = pygame.draw.aalines
        
        draw_line(surf, self.ground.color, self.ground.p1, self.ground.p2)
        draw_lines(surf, self.home.color, 1, self.home.points)
        
        p1,p2 = self.chute.points
        cross_points = line_intersects_rect((p1,p2), self.home.rect, fast=False)
        p1 = cross_points[0]
        draw_line(surf, self.chute.color, p1, p2)
    
    def draw_objects(self):
        surf = State.screen.surface
        draw_aaline = pygame.draw.aaline
        for o in self.space.objects:
            if hasattr(o, 'image'):
                surf.blit(o.image, o.rect)
            else:
                draw_aaline(surf, o.color, o.p1, o.p2)
    
    def draw_achievements(self):
        surf = State.screen.surface
        for sprite in self.score.sprites:
            surf.blit(sprite.image, sprite.rect)
    
    def on_active_event(self, gain, state):
        return
        if gain == 0:
            # Mouse exited pygame window; auto-pause
            if self.do_focus:
                State.clock.pause()
            else:
                # This works around some kind of weird behavior when the
                # window first opens and the mouse position is polled.
                # Sometimes pygame thinks the mouse has left the window
                # area, or started outside the window area or something,
                # firing an ACTIVEEVENT with gain=0. Then you must drag
                # the mouse out and back in to receive gain=1.
                self.do_focus = True
        else:
            # Mouse entered pygame window; auto-unpause
            State.clock.resume()
            pygame.mouse.set_pos(State.camera.screen_center)
            pygame.event.clear()
    
    def paused(self):
        pygame.display.set_caption('Paused')
        pygame.event.set_grab(0)
        pygame.mouse.set_visible(1)
        e = pygame.event.wait()
        if e.type == ACTIVEEVENT:
            if e.gain == 1:
                # Mouse entered pygame window; reset some stuff
                # Need to check if pressed, and set the value appropriately.
                keys = pygame.key.get_pressed()
                self.movex = 0
                for i in settings.KEYS['left']:
                    if keys[i]:
                        self.movex -= State.speed
                for i in settings.KEYS['right']:
                    if keys[i]:
                        self.movex += State.speed
                self.mouse_down_1 = False
                self.mouse_down_3 = False
                State.clock.resume()
                pygame.mouse.set_pos(State.camera.screen_center)
                pygame.event.clear()
                pygame.display.set_caption(settings.caption)
                pygame.event.set_grab(1)
                pygame.mouse.set_visible(0)

#    def on_joy_axis_motion(self, joy, axis, value): pass
#    def on_joy_ball_motion(self, joy, ball, rel): pass
#    def on_joy_button_down(self, joy, button): pass
#    def on_joy_button_up(self, joy, button): pass
#    def on_joy_hat_motion(self, joy, hat, value): pass
    
    def on_key_down(self, unicode, key, mod):
        if key == K_ESCAPE:
            self.on_quit()
        elif key == K_SPACE:
            State.clock.pause()
        elif key == K_RETURN and settings.test_awards:
            self.score.award(5)
        elif key == K_g:
            surf = State.screen.surface
            pygame.image.save(surf, 'screenshot.png')
        elif key == K_m:
            settings.play_music = not settings.play_music
    
#    def on_key_up(self, key, mod): pass
    
    def on_mouse_button_down(self, pos, button):
        self.mouse_buttons[button] = True
    
    def on_mouse_button_up(self, pos, button):
        self.mouse_buttons[button] = False
    
    def on_mouse_motion(self, pos, rel, buttons):
        dx = rel[0]
        angle = self.chute.angle + 90
        if dx > 0:
            if angle + dx > 179:
                angle = 179
            else:
                angle += dx
        elif dx < 0:
            if angle + dx < 1:
                angle = 1
            else:
                angle += dx
        self.chute.angle = angle - 90
    
    def on_quit(self):
        pygame.event.set_grab(0)
        pygame.mouse.set_visible(1)
        sound.song_fadeout(4)
        self.pop()
    
#    def on_user_event(self, e): pass
#    def on_video_expose(self): pass
#    def on_video_resize(self, size, w, h): pass
